/**
 * Example of how to call the TCB DistributedFailureDetection service.
 * 
 *  This server demo is responsible for ending distributed timing failure detections
 * (and the associated measurements) that are started by client demo. 
 * After ending them, it presents the associated measurements results.
 * 
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include sockets header file
#include <winsock.h>
//include tcb services library header file
#include "..\..\..\..\tcb_services_library\tcb_services_library.h"

//TODO: replace this values according to your situation
#define MY_IP "192.168.0.1"
#define MY_PORT 40000
#define MY_MULTICAST_GROUP "224.0.0.10" /* because client sends to an IP multicast */
//TCB event channel
#define EVENT_CHANNEL_ID 10

int main(int argc, char* argv[]) {
	
	//open TCB session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}

	//subscribe to TCB event channel
	error = subscribe_tcb_event_channel(EVENT_CHANNEL_ID);
	switch (error) {
		case 0: printf("Subscribed to EC %d\n", EVENT_CHANNEL_ID); break;
		case -1: printf("Error: Already subscribed to EC %d!\n", EVENT_CHANNEL_ID); break;
		case -2: printf("Error: EC id is invalid!\n"); break;
	}


	//create a socket
	printf("Creating socket\n");
	SOCKET s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s==INVALID_SOCKET) {
		printf("Error creating socket: %d\n", error);
		getchar();
		return -1;
	}

	//bind socket
	printf("Binding socket\n");
	struct sockaddr_in sa;
	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr= inet_addr(MY_IP);
	sa.sin_port=htons(MY_PORT);
	error=bind(s, (struct sockaddr*)&sa, sizeof(sa));
	if (error!=0) {
		printf("Error binding socket: %d\n", error);
		getchar();
		return -1;
	}

	//subscribe to ip multicast group
	struct ip_mreq mreq;
	mreq.imr_multiaddr.s_addr = inet_addr(MY_MULTICAST_GROUP);
	mreq.imr_interface.s_addr = inet_addr(MY_IP);
	error = setsockopt(
     s, 
     IPPROTO_IP,
     IP_ADD_MEMBERSHIP,
     (char*)&mreq, 
     sizeof(mreq));
	if (error!=0) {
		printf("Error joining multicast group: %d\n", error);
		getchar();
		return -1;
	}
	printf("Joined multicast group\n");

	//receive msg infinitely (action_id is in the beginning)
	printf("Waiting data...\n");
	while(1) {
		int buflen = sizeof(tcb_dist_action_id_t)+50;
		char buf[sizeof(tcb_dist_action_id_t)+50];
		struct sockaddr_in from;
		int fromsize = sizeof(from);
		int nbytes = recvfrom(s, buf, buflen, NULL, (struct sockaddr*)&from, 
						  &fromsize);
		if (nbytes<=0) {
			printf("Error receiving data: %d\n", nbytes);
		}
		else {
			tcb_dist_action_id_t action_id;
			tcb_dist_measurement_info_t mi;
			//retrieve action_id from msg first bytes
			memcpy((char*)&action_id, buf, sizeof(action_id));
			//end distributed action (fd and associated measurement) 
			//and get measurement results
			int error = end_tcb_dist_action(&action_id, &mi);
			if (error != 0) {
				printf("Error ending dist action:");
				switch (error) {
				 case -1: printf("no TCB session opened!\n");
						 break;
				 case -2: printf("cannot end distributed FD\n");
						 break;
				 default: printf("unknown\n");
				}
				getchar();
				return -1;
			}
			else {
				//print received msg
				printf("Msg received: %s\n", buf+sizeof(action_id));
				//print measurement results
				printf("Delay: %d ms with error of %d ms\n", mi.delay, mi.error);
			}
		}
	}


	//unsubscribe from TCB event channel
	error = unsubscribe_tcb_event_channel(EVENT_CHANNEL_ID);
	switch (error) {
		case 0: printf("Unsubscribed from EC %d\n", EVENT_CHANNEL_ID); break;
		case -1: printf("Error: Not subscribed to EC %d!\n", EVENT_CHANNEL_ID); break;
		case -2: printf("Error: EC id is invalid!\n"); break;
	}

	//close socket
	printf("Closing socket\n");
	error = closesocket(s);
	if (error!=0) {
		printf("Error closing socket=%d\n",error);
		getchar();
		return -1;
	}
	
	//close TCB session
	printf("Closing TCB session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}
