/**
 * Example of how to call the TCB DistributedRemoteFailureDetection service.
 * 
 * This server demo is responsible for ending the distributed remote timing 
 * failure detections (and the associated measurements) that are started by 
 * client demo. 
 * After ending them, it presents the associated measurements results.
 *
 * A thread is launched to react to timing failures.
 *
 * Before using this demo you need first to create a DLL with
 * two functions:
 * 
 * 		1) The function to timely execute if a failure occurs
 * 		2) A function that returns the Worst Case Execution Time
 * 		   (milliseconds) of the function specified in 1). 
 * 
 * There is already a dll example created that works with this example.
 * The source code (Embedded Visual C++ 3.0) is located in directory
 * "tcb_services_examples/dll_example_for_tcb_handler". You should compile it
 * and deploy in PocketPC prior to running this example.
 * 
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include sockets header file
#include <winsock.h>
//include tcb services library header file
#include "..\..\..\..\tcb_services_library\tcb_services_library.h"

//TODO: replace this values according to your situation
#define MY_IP "192.168.0.3"
#define MY_PORT 40000
#define MY_MULTICAST_GROUP "224.0.0.5"
//TCB event channel
#define EVENT_CHANNEL_ID 5


/**
 * Task that reacts to a timing failure
 */
DWORD WINAPI fail_task (LPVOID param) {
	//create event to receive failure notification
	HANDLE failure_event = CreateEvent(NULL, TRUE, FALSE, L"remote_failure_event");
	if (failure_event==NULL) {
		printf("Error creating failure event\n");
	}
	int count=0;
	while(1) {
		WaitForSingleObject(failure_event, INFINITE);
		ResetEvent(failure_event);
		printf("***failure handler count=%d***\n", count++);
	}
	CloseHandle(failure_event);
	return 0;
}	



int main(int argc, char* argv[]) {
	
	//open_session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}

	//launch fail_task thread
	HANDLE task_id =
		CreateThread(NULL,0,fail_task,NULL,NULL,NULL);
	if (task_id==NULL){
		printf("FATAL ERROR: Couldn't start fail task\n");
		getchar();
		return -1;
	}
	//Give a higher priority to this thread
	CeSetThreadPriority(task_id, 5);

	//define remote FD handler: this handler is executed if a timing failure occurs
	//during a distributed remote failure detection
	printf("Defining remote FD handler\n");
	char dllName[100];
	char funcName[100];
	char wcetName[100];
	//DLL where resides the failure handler and the wcet function
	strcpy(dllName, "tcb_handler_example.dll");
	//failure handler name
	strcpy(funcName, "remote_distFD_example");
	//wcet function name
	strcpy(wcetName, "remote_distFD_example_wcet");
	//maximum time allowed for the execution of failure handler (ms)
	int deadline = 0;
	error=define_tcb_remote_fd_handler(dllName, funcName, wcetName, deadline);
	if (error!=0) {
		printf("Error defining remote FD handler: %d\n", error);
	}

	//subscribe to TCB event channel
	error = subscribe_tcb_event_channel(EVENT_CHANNEL_ID);
	switch (error) {
		case 0: printf("Subscribed to EC %d\n", EVENT_CHANNEL_ID); break;
		case -1: printf("Error: Already subscribed to EC %d!\n", EVENT_CHANNEL_ID); break;
		case -2: printf("Error: EC id is invalid!\n"); break;
	}

	
	//create a socket
	printf("Creating socket\n");
	SOCKET s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s==INVALID_SOCKET) {
		printf("Error creating socket: %d\n", error);
		getchar();
		return -1;
	}

	//bind socket
	printf("Binding socket\n");
	struct sockaddr_in sa;
	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr= inet_addr(MY_IP);
	sa.sin_port=htons(MY_PORT);
	error=bind(s, (struct sockaddr*)&sa, sizeof(sa));
	if (error!=0) {
		printf("Error binding socket: %d\n", error);
		getchar();
		return -1;
	}

	//subscribe to ip multicast group
	struct ip_mreq mreq;
	mreq.imr_multiaddr.s_addr = inet_addr(MY_MULTICAST_GROUP);
	mreq.imr_interface.s_addr = inet_addr(MY_IP);
	error = setsockopt(
     s, 
     IPPROTO_IP,
     IP_ADD_MEMBERSHIP,
     (char*)&mreq, 
     sizeof(mreq));
	if (error!=0) {
		printf("Error joining multicast group: %d\n", error);
		getchar();
		return -1;
	}
	printf("Joined multicast group\n");


	//receive msg infinitely (action_id is in the beginning)
	printf("Waiting data...\n");
	while (1) {
		int buflen = sizeof(tcb_dist_action_id_t)+50;
		char buf[sizeof(tcb_dist_action_id_t)+50];
		struct sockaddr_in from;
		int fromsize = sizeof(from);
		int nbytes = recvfrom(s, buf, buflen, NULL, (struct sockaddr*)&from, 
						  &fromsize);
		if (nbytes<=0) {
			printf("Error receiving data: %d\n", nbytes);
		}
		else {
			tcb_dist_action_id_t action_id;
			tcb_dist_measurement_info_t mi;
			//retrieve action_id from msg first bytes
			memcpy((char*)&action_id, buf, sizeof(action_id));
			//end distributed action (remote fd and associated measurement) 
			//and get measurement results
			int error = end_tcb_dist_action(&action_id, &mi);
			if (error != 0) {
				printf("Error ending dist action:");
				switch (error) {
					case -1: printf("no TCB session opened!\n");
						 break;
					case -2: printf("cannot end distributed FD\n");
						 break;
					default: printf("unknown\n");
				}
				getchar();
				return -1;
			}
			else {
				//print received msg
				printf("Msg received: %s\n", buf+sizeof(action_id));
				//print measurement results
				printf("Delay: %d ms with error of %d ms\n", mi.delay, mi.error);
			}
		}
	}

	//unsubscribe from TCB event channel
	error = unsubscribe_tcb_event_channel(EVENT_CHANNEL_ID);
	switch (error) {
		case 0: printf("Unsubscribed from EC %d\n", EVENT_CHANNEL_ID); break;
		case -1: printf("Error: Not subscribed to EC %d!\n", EVENT_CHANNEL_ID); break;
		case -2: printf("Error: EC id is invalid!\n"); break;
	}

	//close socket
	printf("Closing socket\n");
	error = closesocket(s);
	if (error!=0) {
		printf("Error closing socket=%d\n",error);
		getchar();
		return -1;
	}
	
	//close TCB session
	printf("Closing TCB session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}
