/**
 * Header file of TCB services C library.
 *
 * author: Paulo Sousa (pjsousa@di.fc.ul.pt)
 * date: 21/03/2003
 */


// The following ifdef block is the standard way of creating macros which make exporting 
// from a DLL simpler. All files within this DLL are compiled with the TCB_SERVICES_LIBRARY_EXPORTS
// symbol defined on the command line. this symbol should not be defined on any project
// that uses this DLL. This way any other project whose source files include this file see 
// TCB_SERVICES_LIBRARY_API functions as being imported from a DLL, wheras this DLL sees symbols
// defined with this macro as being exported.
#ifdef TCB_SERVICES_LIBRARY_EXPORTS
#define TCB_SERVICES_LIBRARY_API __declspec(dllexport)
#else
#define TCB_SERVICES_LIBRARY_API __declspec(dllimport)
#endif


/**
 * Only used for tests
 */
extern TCB_SERVICES_LIBRARY_API int nTcb_services_library;
TCB_SERVICES_LIBRARY_API int fnTcb_services_library(void);

//we need this for MAX_NUMBER_OF_MACHINES
#include "../../tcb_wince/config_parameters.h"

/* NOTE: All time intervals are presented/should be specified in milliseconds. */


/**
 * Info returned by the local measurement service
 */
typedef struct tcb_measurement_info {
	//end instant
	int end_ev;
	//measurement value
	int procdelay;
} tcb_measurement_info_t;


/**
 * Info returned by the local failure detection service
 */
typedef struct tcb_local_fd_info {
	//end instant
	int end_ev;
	//time elapsed between failure detection start instant and end_ev
	int procdelay;
	//indicates if a timing failure occured
	bool faulty;
} tcb_local_fd_info_t;


/**
 * Info about the result of a distributed timing failure detection in a machine.
 */
typedef struct tcb_distributed_fail_info {
	bool empty;
	//TCB id
	unsigned long id;
	//indicates if a failure occured:
	//0-no failure
	//1-failure
	//2-no info
	char fail;
	//indicates if lateness is available
	bool latenessAvail;
	//indicates the lateness degree if it's available
	int latenessDegree;
} tcb_distributed_fail_info_t;


/**
 * Info returned by wait_info. It contains information about the result of
 * a distributed timing failure detection.
 */
typedef struct tcb_distributed_fd_info {
	//information sender id
	unsigned long from;
	//information sequence number
	int seq;
	//info about failure in each machine
	tcb_distributed_fail_info_t failInfo[MAX_NUMBER_OF_MACHINES];
} tcb_distributed_fd_info_t;


/**
 * Info returned by end_tcb_dist_measurement and end_tcb_dist_fd
 * It contains information about the result of a distributed measurement.
 */
typedef struct tcb_distributed_measurement_info {
	//instant of termination
	int end_ev;
	//duration of the action terminated at end_ev
	int delay;
	//measurement error (ms)
	int error;
} tcb_dist_measurement_info_t;


//stores info about a distributed action: measurement, fd, remote_fd
typedef struct tcb_dist_action_id {
	//0=measurement, >0=fd or remote fd
	int seq;
	//global timestamp at start_ev
	int start_ev_global;
	//sender tcb id
	unsigned long tcb_id;
} tcb_dist_action_id_t;


/**
 * Opens the TCB session. This method should be called
 * first of all. 
 * In this version, only 1 session is allowed per machine.
 * 
 * @return 0 if OK, -1 if TCB is stopped, 
 *					-2 if a session is already open,
 *					-3 if other error
 */
TCB_SERVICES_LIBRARY_API int open_tcb_session(void);

/**
 * Closes the TCB session.
 * 
 * @return 0 if OK, -1 if error
 */
TCB_SERVICES_LIBRARY_API int close_tcb_session(void);


/**
 * Calls the tcb service: subscribe_event_channel
 *
 * Subscribe to a certain event channel.
 *
 * @param ec_id Id of the channel to subscribe (0<=ec_id<=31)
 *
 * @return 0 if OK,
 *		   -1 if already subscribed,
 *		   -2 if ec_id is invalid.
 */
TCB_SERVICES_LIBRARY_API int subscribe_tcb_event_channel(char ec_id);


/**
 * Calls the tcb service: unsubscribe_event_channel
 *
 * Unsubscribe from a certain event channel.
 *
 * @param ec_id Id of the channel to unsubscribe (0<=ec_id<=31)
 *
 * @return 0 if OK,
 *		   -1 if not subscribed,
 *		   -2 if ec_id is invalid.
 */
TCB_SERVICES_LIBRARY_API int unsubscribe_tcb_event_channel(char ec_id);


/**
 * Calls the tcb service: get_timestamp
 * 
 * Returns the local clock value (milliseconds).
 * 
 * @return Local clock value (milliseconds) if OK,
 *		   -1 if no TCB session is opened		   
 */
TCB_SERVICES_LIBRARY_API int get_tcb_timestamp(void);

/**
 * Calls the tcb service: get_global_timestamp
 * 
 * Returns the global clock value (milliseconds) at local instant specified.
 * If localInstant is 0, the present global clock value is returned.
 * 
 * @param localInstant A timestamp obtained with get_timestamp() or 0.
 * 
 * @return Global clock value (milliseconds) at local instant specified if OK
 * 		   (If localInstant is 0, the present global clock value is returned),
 *			-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int get_tcb_global_timestamp(int localInstant);

/**
 * Calls the tcb pseudo-service: get_tcb_id
 * 
 * Returns the id of the TCB we are connected to.
 * 
 * @return Id of the TCB we are connected to if OK,
 *		   -1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API unsigned long get_tcb_id(void);

/**
 * Calls the tcb service: start_local_measurement
 * 
 * Starts a local measurement at instant start_ev.
 * 
 * @param start_ev Instant that marks the begin of the measurement
 * 
 * @return Tag that identifies the measurement if OK,
 *		   -1 if no tag is available,
 *		   -2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_measurement(int start_ev);

/**
 * Calls the tcb service: end_local_measurement
 * 
 * Stops the measurement identified by tag.
 * 
 * @param tag Measurement identifier
 *
 * @return Information about measurement if OK,
 *		   tcb_measurement_info_t.end_ev=-1 if tag is invalid,
 *		   tcb_measurement_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_measurement_info_t end_tcb_local_measurement(int tag);

/**
 * Calls the tcb service: timely_exec
 * 
 * Starts the execution of an application function. 
 * TCB verifies if it's possible to timely execute it.
 * 
 * @param start_ev A time instant
 * @param delay start_ev+delay marks the exec begin
 * @param max_exec Deadline for the execution, counting since start_ev
 * @param dllName Name of the DLL file where the function is defined
 * @param funcName Function name 
 * @param wcetName Worst Case Execution Time function name
 * 
 * @return The time instant when the function ends if OK;
 *			-1 if no available entry in the executions table
 *			-2 if dll not found
 *			-3 if function not found
 *			-4 if wcet function not found
 *			-5 if it's not possible to execute function
 *			-6 if it's not possible to timely execute function
 *			-7 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int tcb_timely_exec(int start_ev, int delay, int max_exec, char* dllName,
										 char* funcName, char* wcetName);

/**
 * Calls the tcb service: start_local_fd
 *
 * Starts a local timing failure detection.
 * 
 * @param start_ev A time instant
 * @param spec A timing failure occurs if start_ev+spec instant is reached
 *	            before a call endLocalFD
 * @param deadline Deadline for the execution of handler, if a failure occurs
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 *
 * @return Tag that identifies the timing failure detection if OK,
 *			-1 if no available entry in the localFD table,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if it's not possible to execute timing failure detection,
 *			-6 if a timing failure already occured,
 *			-7 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_fd(int start_ev, int spec, int deadline, 
											char* dllName, char* funcName, char* wcetName);

/**
 * Calls the tcb service: end_local_fd
 *
 * Ends a local timing failure detection.
 * 
 * @param tag Local timing detection identifier
 *
 * @return Information about the timing failure detection if OK,
 *			local_fd_info_t.end_ev=-1 if tag is invalid,
 *			local_fd_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_local_fd_info_t end_tcb_local_fd(int tag);

/**
 * Waits for info from TCB distributed timing failure detection service.
 *
 * @return Information about the distributed timing failure detection if OK,
 *		   tcb_distributed_fd_info_t.from=-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_fd_info_t tcb_wait_info(void);


/**
 * Calls the TCB pseudo-service: define_remote_fd_handler
 *
 * Defines the handler to be used on remote failure detection.
 *
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 * @param deadline Maximum exec time of failure handler (if a failure occurs)
 *
 * @return 0 if OK,
 *		-2 if dll not found,
 *		-3 if function not found,
 *		-4 if wcet function not found,
 *		-5 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int define_tcb_remote_fd_handler(char* dllName, char* funcName, char* wcetName, 
							 int deadline);


/**
 * Calls the TCB service: start_dist_measurement
 * 
 * Starts a distributed measurement at instant start_ev.
 * 
 * @param IN start_ev Instant that marks the begin of the measurement.
 * @param OUT action_id Id that identifies the distributed action.
 * 
 * @return  0 if OK,
 *		   -1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API 
int start_tcb_dist_measurement(int start_ev, tcb_dist_action_id_t *action_id);



/**
 * Calls the TCB service: start_dist_fd
 * 
 * Starts simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the sender.
 *
 * If detection is not ended in all recipients subscribed to event_channel_id 
 * until start_ev+spec,
 * the specified failure handler is executed (in the sender) with the maximum
 * exec time specified.
 *
 * @param IN start_ev the initial time instant of the failure detection and measurement
 * @param IN event_channel_id Event channel where the timing failure detection will be
 *						   started (0<=event_channel_id<=31)
 * @param IN deadline Maximum exec time of failure handler (if a failure occurs)
 * @param IN dllName Name of the DLL file where the failure handler is defined
 * @param IN funcName Failure handler name
 * @param IN wcetName Failure handler Worst Case Execution Time function name
 * @param OUT action_id Id that identifies the distributed action.
 *
 * @return 0 if OK, 
 *		   -1 if no TCB session is opened,
 *		   -2 if dll not found,
 *		   -3 if function not found,
 *		   -4 if wcet function not found,
 *		   -5 if there are too many distributed fds already running,
 *		   -6 if a timing failure already occured,
 *		   -7 if it's not possible to execute timing failure detection
 */
TCB_SERVICES_LIBRARY_API 
int start_tcb_dist_fd(int start_ev, char event_channel_id, int spec, int deadline,
					  char* dllName, char* funcName, char* wcetName, 
					  tcb_dist_action_id_t *action_id);


/**
 * Calls the TCB service: start_dist_remote_fd
 * 
 * Starts simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the recipients.
 *
 * For each recipient specified in dest_mask: if service is not ended until start_ev+spec,
 * the failure handler specified in define_tcb_remote_fd_handler
 * is executed with the maximum exec time specified in that function.
 *
 * @param IN start_ev the initial time instant of the failure detection and measurement
 * @param IN dest_mask bit mask indicating service destinations. If bit n=1,
 *				    then service is sent to destination where TCB id=n.
 *				    E.g. dest_mask=0x7 => send to destinations with TCB ids=0,1,2.
 * @param IN spec Maximum duration starting at start_ev (for failure detection)
 * @param OUT action_id Id that identifies the distributed action.
 *
 * @return 0 if OK, 
 *		   -1 if no TCB session is opened,
 *		   -5 if there are too many distributed fds already running,
 *		   -6 if a timing failure already occured,
 *		   -7 if it's not possible to execute timing failure detection
 */
TCB_SERVICES_LIBRARY_API 
int start_tcb_dist_remote_fd(int start_ev, char dest_mask, int spec, 
							 tcb_dist_action_id_t *action_id);



/**
 * Calls the tcb service: end_dist_action
 * 
 * Ends the distributed action (measurement/fd/remote fd)
 * identified by action_id.
 * 
 * @param IN action_id Distributed action id.
 * @param OUT measurement_info Info about the distributed measurement result.
 *
 * @return 0 if OK,
 *		   -1 if no TCB session is opened,
 *		   -2 if cannot end distributed FD.
 */
TCB_SERVICES_LIBRARY_API 
int end_tcb_dist_action(tcb_dist_action_id_t *action_id, 
						tcb_dist_measurement_info_t *measurement_info);